@@ -1,5 +1,5 @@
 ########
-# Copyright (c) 2014 GigaSpaces Technologies Ltd. All rights reserved
+# Copyright (c) 2018 Cloudify Platform Ltd. All rights reserved
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,39 +14,42 @@
 # limitations under the License.
 ############
 
-
 import os
 import json
 import shutil
-from urlparse import urlparse
 
 import click
 
 from dsl_parser.parser import parse_from_path
 from dsl_parser.exceptions import DSLParsingException
+from cloudify._compat import urlparse
 from cloudify_rest_client.constants import VISIBILITY_EXCEPT_PRIVATE
 
 from .. import local
 from .. import utils
-from ..cli import cfy
+from ..cli import cfy, helptexts
 from .. import blueprint
 from .. import exceptions
 from ..config import config
-from ..table import print_data
+from ..logger import get_global_json_output
+from ..table import print_data, print_single
 from ..exceptions import CloudifyCliError
 from ..utils import (prettify_client_error,
                      get_visibility,
                      validate_visibility)
+from .summary import BASE_SUMMARY_FIELDS, structure_summary_results
 
 
 DESCRIPTION_LIMIT = 20
-BLUEPRINT_COLUMNS = ['id', 'description', 'main_file_name', 'created_at',
-                     'updated_at', 'visibility', 'tenant_name', 'created_by']
+BASE_BLUEPRINT_COLUMNS = ['id', 'description', 'main_file_name', 'created_at']
+BLUEPRINT_COLUMNS = BASE_BLUEPRINT_COLUMNS + ['updated_at', 'visibility',
+                                              'tenant_name', 'created_by']
 INPUTS_COLUMNS = ['name', 'type', 'default', 'description']
+BLUEPRINTS_SUMMARY_FIELDS = BASE_SUMMARY_FIELDS
 
 
 @cfy.group(name='blueprints')
-@cfy.options.verbose()
+@cfy.options.common_options
 def blueprints():
     """Handle blueprints on the manager
     """
@@ -56,7 +59,7 @@ def blueprints():
 @blueprints.command(name='validate',
                     short_help='Validate a blueprint')
 @cfy.argument('blueprint-path')
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.pass_logger
 def validate_blueprint(blueprint_path, logger):
     """Validate a blueprint
@@ -72,17 +75,17 @@ def validate_blueprint(blueprint_path, logger):
             resolver=resolver,
             validate_version=validate_version)
     except DSLParsingException as ex:
-        raise CloudifyCliError('Failed to validate blueprint {0}'.format(ex))
+        raise CloudifyCliError('Failed to validate blueprint: {0}'.format(ex))
     logger.info('Blueprint validated successfully')
 
 
 @blueprints.command(name='upload',
                     short_help='Upload a blueprint [manager only]')
 @cfy.argument('blueprint-path')
-@cfy.options.blueprint_id()
+@cfy.options.blueprint_id(validate=True)
 @cfy.options.blueprint_filename()
 @cfy.options.validate
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.options.tenant_name(required=False, resource_name_for_help='blueprint')
 @cfy.options.private_resource
 @cfy.options.visibility()
@@ -108,15 +111,19 @@ def upload(ctx,
     retrieved from GitHub).
     Supported archive types are: zip, tar, tar.gz and tar.bz2
     """
-    if tenant_name:
-        logger.info('Explicitly using tenant `{0}`'.format(tenant_name))
-
+    utils.explicit_tenant_name_message(tenant_name, logger)
     processed_blueprint_path = blueprint.get(
         blueprint_path, blueprint_filename)
 
     # Take into account that `blueprint.get` might not return a URL
     # instead of a blueprint file (archive files are not locally downloaded)
-    is_url = bool(urlparse(processed_blueprint_path).scheme)
+    parsed_path = urlparse(processed_blueprint_path)
+    is_url = bool(
+        parsed_path.scheme
+        # But on windows there will always be a scheme, so check if the path
+        # exists
+        and not os.path.exists(processed_blueprint_path)
+    )
 
     progress_handler = utils.generate_progress_handler(blueprint_path, '')
     blueprint_id = blueprint_id or blueprint.generate_id(
@@ -150,7 +157,9 @@ def upload(ctx,
                 processed_blueprint_path,
                 blueprint_id,
                 visibility,
-                progress_handler
+                progress_handler,
+                # if blueprint is in an archive we skip the size limit check
+                utils.is_archive(blueprint_path)
             )
         finally:
             # When an archive file is passed, it's extracted to a temporary
@@ -170,7 +179,7 @@ def upload(ctx,
                     short_help='Download a blueprint [manager only]')
 @cfy.argument('blueprint-id')
 @cfy.options.output_path
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.options.tenant_name(required=False, resource_name_for_help='blueprint')
 @cfy.assert_manager_active()
 @cfy.pass_client()
@@ -180,8 +189,7 @@ def download(blueprint_id, output_path, logger, client, tenant_name):
 
     `BLUEPRINT_ID` is the id of the blueprint to download.
     """
-    if tenant_name:
-        logger.info('Explicitly using tenant `{0}`'.format(tenant_name))
+    utils.explicit_tenant_name_message(tenant_name, logger)
     logger.info('Downloading blueprint {0}...'.format(blueprint_id))
     blueprint_name = output_path if output_path else blueprint_id
     progress_handler = utils.generate_progress_handler(blueprint_name, '')
@@ -194,44 +202,45 @@ def download(blueprint_id, output_path, logger, client, tenant_name):
 @blueprints.command(name='delete',
                     short_help='Delete a blueprint [manager only]')
 @cfy.argument('blueprint-id')
-@cfy.options.verbose()
+@cfy.options.force(help=helptexts.FORCE_DELETE_BLUEPRINT)
+@cfy.options.common_options
 @cfy.options.tenant_name(required=False, resource_name_for_help='blueprint')
 @cfy.assert_manager_active()
 @cfy.pass_client()
 @cfy.pass_logger
-def delete(blueprint_id, logger, client, tenant_name):
+def delete(blueprint_id, force, logger, client, tenant_name):
     """Delete a blueprint from the manager
 
     `BLUEPRINT_ID` is the id of the blueprint to delete.
     """
-    if tenant_name:
-        logger.info('Explicitly using tenant `{0}`'.format(tenant_name))
+    utils.explicit_tenant_name_message(tenant_name, logger)
     logger.info('Deleting blueprint {0}...'.format(blueprint_id))
-    client.blueprints.delete(blueprint_id)
+    client.blueprints.delete(blueprint_id, force)
     logger.info('Blueprint deleted')
 
 
-@blueprints.command(name='list',
-                    short_help='List blueprints [manager only]')
+@cfy.command(name='list', short_help='List blueprints')
 @cfy.options.sort_by()
 @cfy.options.descending
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.options.tenant_name_for_list(
     required=False, resource_name_for_help='blueprint')
 @cfy.options.all_tenants
+@cfy.options.search
 @cfy.options.pagination_offset
 @cfy.options.pagination_size
 @cfy.assert_manager_active()
 @cfy.pass_client()
 @cfy.pass_logger
-def list(sort_by,
-         descending,
-         tenant_name,
-         all_tenants,
-         pagination_offset,
-         pagination_size,
-         logger,
-         client):
+def manager_list(sort_by,
+                 descending,
+                 tenant_name,
+                 all_tenants,
+                 search,
+                 pagination_offset,
+                 pagination_size,
+                 logger,
+                 client):
     """List all blueprints
     """
     def trim_description(blueprint):
@@ -243,13 +252,13 @@ def list(sort_by,
             blueprint['description'] = ''
         return blueprint
 
-    if tenant_name:
-        logger.info('Explicitly using tenant `{0}`'.format(tenant_name))
+    utils.explicit_tenant_name_message(tenant_name, logger)
     logger.info('Listing all blueprints...')
     blueprints_list = client.blueprints.list(
         sort=sort_by,
         is_descending=descending,
         _all_tenants=all_tenants,
+        _search=search,
         _offset=pagination_offset,
         _size=pagination_size
     )
@@ -259,10 +268,17 @@ def list(sort_by,
     logger.info('Showing {0} of {1} blueprints'.format(len(blueprints), total))
 
 
+@cfy.command(name='list', short_help='List blueprints')
+@cfy.pass_logger
+def local_list(logger):
+    blueprints = local.list_blueprints()
+    print_data(BASE_BLUEPRINT_COLUMNS, blueprints, 'Blueprints:')
+
+
 @blueprints.command(name='get',
                     short_help='Retrieve blueprint information [manager only]')
 @cfy.argument('blueprint-id')
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.options.tenant_name(required=False, resource_name_for_help='blueprint')
 @cfy.assert_manager_active()
 @cfy.pass_client()
@@ -272,27 +288,42 @@ def get(blueprint_id, logger, client, tenant_name):
 
     `BLUEPRINT_ID` is the id of the blueprint to get information on.
     """
-    if tenant_name:
-        logger.info('Explicitly using tenant `{0}`'.format(tenant_name))
+    utils.explicit_tenant_name_message(tenant_name, logger)
     logger.info('Retrieving blueprint {0}...'.format(blueprint_id))
     blueprint_dict = client.blueprints.get(blueprint_id)
     deployments = client.deployments.list(_include=['id'],
                                           blueprint_id=blueprint_id)
     blueprint_dict['#deployments'] = len(deployments)
     columns = BLUEPRINT_COLUMNS + ['#deployments']
-    print_data(columns, blueprint_dict, 'Blueprint:', max_width=50)
+    blueprint_metadata = blueprint_dict['plan']['metadata'] or {}
+    blueprint_deployments = [d['id'] for d in deployments]
 
-    logger.info('Description:')
-    logger.info('{0}\n'.format(blueprint_dict['description'] or ''))
+    if get_global_json_output():
+        columns += ['description', 'metadata', 'deployments']
+        blueprint_dict['metadata'] = blueprint_metadata
+        blueprint_dict['deployments'] = blueprint_deployments
+        print_single(columns, blueprint_dict, 'Blueprint:', max_width=50)
+    else:
+        print_single(columns, blueprint_dict, 'Blueprint:', max_width=50)
 
-    logger.info('Existing deployments:')
-    logger.info('{0}\n'.format(json.dumps([d['id'] for d in deployments])))
+        logger.info('Description:')
+        logger.info('{0}\n'.format(blueprint_dict['description'] or ''))
+
+        if blueprint_metadata:
+            logger.info('Metadata:')
+            for property_name, property_value in \
+                    blueprint_dict['plan']['metadata'].items():
+                logger.info('\t{0}: {1}'.format(property_name, property_value))
+            logger.info('')
+
+        logger.info('Existing deployments:')
+        logger.info('{0}\n'.format(json.dumps(blueprint_deployments)))
 
 
 @blueprints.command(name='inputs',
                     short_help='Retrieve blueprint inputs [manager only]')
 @cfy.argument('blueprint-id')
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.options.tenant_name(required=False, resource_name_for_help='blueprint')
 @cfy.assert_manager_active()
 @cfy.pass_client()
@@ -302,8 +333,7 @@ def inputs(blueprint_id, logger, client, tenant_name):
 
     `BLUEPRINT_ID` is the path of the blueprint to get inputs for.
     """
-    if tenant_name:
-        logger.info('Explicitly using tenant `{0}`'.format(tenant_name))
+    utils.explicit_tenant_name_message(tenant_name, logger)
     logger.info('Retrieving inputs for blueprint {0}...'.format(blueprint_id))
     blueprint_dict = client.blueprints.get(blueprint_id)
     inputs = blueprint_dict['plan']['inputs']
@@ -311,7 +341,7 @@ def inputs(blueprint_id, logger, client, tenant_name):
              'type': input.get('type', '-'),
              'default': input.get('default', '-'),
              'description': input.get('description', '-')}
-            for name, input in inputs.iteritems()]
+            for name, input in inputs.items()]
 
     print_data(INPUTS_COLUMNS, data, 'Inputs:')
 
@@ -321,7 +351,7 @@ def inputs(blueprint_id, logger, client, tenant_name):
 @cfy.argument('blueprint-path')
 @cfy.options.optional_output_path
 @cfy.options.validate
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.pass_logger
 @cfy.pass_context
 def package(ctx, blueprint_path, output_path, validate, logger):
@@ -345,7 +375,7 @@ def package(ctx, blueprint_path, output_path, validate, logger):
             "You must provide a path to a blueprint's directory or to a "
             "blueprint yaml file residing in a blueprint's directory.")
     if os.name == 'nt':
-        utils.zip(path_to_package, destination + '.zip')
+        utils.create_zip(path_to_package, destination + '.zip')
     else:
         utils.tar(path_to_package, destination + '.tar.gz')
     logger.info('Packaging complete!')
@@ -355,7 +385,7 @@ def package(ctx, blueprint_path, output_path, validate, logger):
                     short_help='Create pip-requirements')
 @cfy.argument('blueprint-path', type=click.Path(exists=True))
 @cfy.options.optional_output_path
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.pass_logger
 def create_requirements(blueprint_path, output_path, logger):
     """Generate a pip-compliant requirements file for a given blueprint
@@ -381,7 +411,7 @@ def create_requirements(blueprint_path, output_path, logger):
 @blueprints.command(name='install-plugins',
                     short_help='Install plugins [locally]')
 @cfy.argument('blueprint-path', type=click.Path(exists=True))
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.assert_local_active
 @cfy.pass_logger
 def install_plugins(blueprint_path, logger):
@@ -399,7 +429,7 @@ def install_plugins(blueprint_path, logger):
 @blueprints.command(name='set-global',
                     short_help="Set the blueprint's visibility to global")
 @cfy.argument('blueprint-id')
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.assert_manager_active()
 @cfy.pass_client(use_tenant_in_header=True)
 @cfy.pass_logger
@@ -420,7 +450,7 @@ def set_global(blueprint_id, logger, client):
                     short_help="Set the blueprint's visibility")
 @cfy.argument('blueprint-id')
 @cfy.options.visibility(required=True, valid_values=VISIBILITY_EXCEPT_PRIVATE)
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.assert_manager_active()
 @cfy.pass_client(use_tenant_in_header=True)
 @cfy.pass_logger
@@ -435,3 +465,45 @@ def set_visibility(blueprint_id, visibility, logger, client):
         client.blueprints.set_visibility(blueprint_id, visibility)
         logger.info('Blueprint `{0}` was set to {1}'.format(blueprint_id,
                                                             visibility))
+
+
+@blueprints.command(name='summary',
+                    short_help='Retrieve summary of blueprint details '
+                               '[manager only]')
+@cfy.argument('target_field', type=click.Choice(BLUEPRINTS_SUMMARY_FIELDS))
+@cfy.argument('sub_field', type=click.Choice(BLUEPRINTS_SUMMARY_FIELDS),
+              default=None, required=False)
+@cfy.options.common_options
+@cfy.options.tenant_name(required=False, resource_name_for_help='summary')
+@cfy.options.all_tenants
+@cfy.pass_logger
+@cfy.pass_client()
+def summary(target_field, sub_field, logger, client, tenant_name,
+            all_tenants):
+    """Retrieve summary of blueprints, e.g. a count of each blueprint with
+    the same tenant name.
+
+    `TARGET_FIELD` is the field to summarise blueprints on.
+    """
+    utils.explicit_tenant_name_message(tenant_name, logger)
+    logger.info('Retrieving summary of blueprints on field {field}'.format(
+        field=target_field))
+
+    summary = client.summary.blueprints.get(
+        _target_field=target_field,
+        _sub_field=sub_field,
+        _all_tenants=all_tenants,
+    )
+
+    columns, items = structure_summary_results(
+        summary.items,
+        target_field,
+        sub_field,
+        'blueprints',
+    )
+
+    print_data(
+        columns,
+        items,
+        'Blueprint summary by {field}'.format(field=target_field),
+    )
