@@ -1,5 +1,5 @@
 ########
-# Copyright (c) 2014 GigaSpaces Technologies Ltd. All rights reserved
+# Copyright (c) 2014-2019 Cloudify Platform Ltd. All rights reserved
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -17,29 +17,46 @@
 import json
 import time
 
+import click
 from cloudify_rest_client import exceptions
 
 from .. import local
 from .. import utils
-from ..table import print_data
+from ..table import print_data, print_single, print_details
+from ..utils import get_deployment_environment_execution
 from ..cli import cfy, helptexts
-from ..logger import get_events_logger
-from ..constants import DEFAULT_UNINSTALL_WORKFLOW
+from ..logger import get_events_logger, get_global_json_output
+from ..constants import DEFAULT_UNINSTALL_WORKFLOW, CREATE_DEPLOYMENT
 from ..execution_events_fetcher import wait_for_execution
 from ..exceptions import CloudifyCliError, ExecutionTimeoutError, \
     SuppressedCloudifyCliError
+from .summary import BASE_SUMMARY_FIELDS, structure_summary_results
 
 _STATUS_CANCELING_MESSAGE = (
     'NOTE: Executions currently in a "canceling/force-canceling" status '
     'may take a while to change into "cancelled"')
 
-EXECUTION_COLUMNS = ['id', 'workflow_id', 'status', 'deployment_id',
-                     'created_at', 'error', 'visibility', 'tenant_name',
-                     'created_by']
+BASE_EXECUTION_COLUMNS = ['id', 'workflow_id', 'status_display']
+LOCAL_EXECUTION_COLUMNS = BASE_EXECUTION_COLUMNS + [
+    'blueprint_id', 'started_at', 'ended_at', 'error']
+FULL_EXECUTION_COLUMNS = BASE_EXECUTION_COLUMNS + [
+    'is_dry_run', 'deployment_id', 'blueprint_id', 'created_at', 'ended_at',
+    'error', 'visibility', 'tenant_name', 'created_by', 'started_at',
+    'scheduled_for']
+MINIMAL_EXECUTION_COLUMNS = BASE_EXECUTION_COLUMNS + [
+    'is_dry_run', 'deployment_id', 'created_at', 'started_at', 'scheduled_for',
+    'visibility', 'tenant_name', 'created_by']
+EXECUTION_TABLE_LABELS = {'status_display': 'status'}
+EXECUTIONS_SUMMARY_FIELDS = [
+    'status',
+    'blueprint_id',
+    'deployment_id',
+    'workflow_id',
+] + BASE_SUMMARY_FIELDS
 
 
 @cfy.group(name='executions')
-@cfy.options.verbose()
+@cfy.options.common_options
 def executions():
     """Handle workflow executions
     """
@@ -49,7 +66,7 @@ def executions():
 @cfy.command(name='get',
              short_help='Retrieve execution information [manager only]')
 @cfy.argument('execution-id')
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.options.tenant_name(required=False, resource_name_for_help='execution')
 @cfy.assert_manager_active()
 @cfy.pass_client()
@@ -59,8 +76,7 @@ def manager_get(execution_id, logger, client, tenant_name):
 
     `EXECUTION_ID` is the execution to get information on.
     """
-    if tenant_name:
-        logger.info('Explicitly using tenant `{0}`'.format(tenant_name))
+    utils.explicit_tenant_name_message(tenant_name, logger)
     try:
         logger.info('Retrieving execution {0}'.format(execution_id))
         execution = client.executions.get(execution_id)
@@ -69,20 +85,21 @@ def manager_get(execution_id, logger, client, tenant_name):
             raise
         raise CloudifyCliError('Execution {0} not found'.format(execution_id))
 
-    print_data(EXECUTION_COLUMNS, execution, 'Execution:', max_width=50)
+    columns = FULL_EXECUTION_COLUMNS
+    if get_global_json_output():
+        columns += ['parameters']
+    print_single(columns, execution, 'Execution:', max_width=50,
+                 labels=EXECUTION_TABLE_LABELS)
 
-    # print execution parameters
-    logger.info('Execution Parameters:')
-    for param_name, param_value in utils.decode_dict(
-            execution.parameters).iteritems():
-        logger.info('\t{0}: \t{1}'.format(param_name, param_value))
+    if not get_global_json_output():
+        print_details(execution.parameters, 'Execution Parameters:')
     if execution.status in (execution.CANCELLING, execution.FORCE_CANCELLING):
         logger.info(_STATUS_CANCELING_MESSAGE)
     logger.info('')
 
 
 @cfy.command(name='list',
-             short_help='List deployment executions [manager only]')
+             short_help='List deployment executions')
 @cfy.options.deployment_id(required=False)
 @cfy.options.include_system_workflows
 @cfy.options.sort_by()
@@ -92,7 +109,7 @@ def manager_get(execution_id, logger, client, tenant_name):
 @cfy.options.all_tenants
 @cfy.options.pagination_offset
 @cfy.options.pagination_size
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.assert_manager_active()
 @cfy.pass_client()
 @cfy.pass_logger
@@ -112,8 +129,7 @@ def manager_list(
     If `DEPLOYMENT_ID` is provided, list executions for that deployment.
     Otherwise, list executions for all deployments.
     """
-    if tenant_name:
-        logger.info('Explicitly using tenant `{0}`'.format(tenant_name))
+    utils.explicit_tenant_name_message(tenant_name, logger)
     try:
         if deployment_id:
             logger.info('Listing executions for deployment {0}...'.format(
@@ -136,7 +152,8 @@ def manager_list(
         raise CloudifyCliError('Deployment {0} does not exist'.format(
             deployment_id))
 
-    print_data(EXECUTION_COLUMNS, executions, 'Executions:')
+    print_data(MINIMAL_EXECUTION_COLUMNS, executions, 'Executions:',
+               labels=EXECUTION_TABLE_LABELS)
     total = executions.metadata.pagination.total
     logger.info('Showing {0} of {1} executions'.format(len(executions), total))
 
@@ -147,7 +164,7 @@ def manager_list(
 
 
 @cfy.command(name='start',
-             short_help='Execute a workflow [manager only]')
+             short_help='Execute a workflow')
 @cfy.argument('workflow-id')
 @cfy.options.deployment_id(required=True)
 @cfy.options.parameters
@@ -156,8 +173,12 @@ def manager_list(
 @cfy.options.timeout()
 @cfy.options.include_logs
 @cfy.options.json_output
-@cfy.options.verbose()
+@cfy.options.dry_run
+@cfy.options.wait_after_fail
+@cfy.options.common_options
 @cfy.options.tenant_name(required=False, resource_name_for_help='execution')
+@cfy.options.schedule
+@cfy.options.queue
 @cfy.assert_manager_active()
 @cfy.pass_client()
 @cfy.pass_logger
@@ -169,6 +190,10 @@ def manager_start(workflow_id,
                   timeout,
                   include_logs,
                   json_output,
+                  dry_run,
+                  wait_after_fail,
+                  queue,
+                  schedule,
                   logger,
                   client,
                   tenant_name):
@@ -176,17 +201,15 @@ def manager_start(workflow_id,
 
     `WORKFLOW_ID` is the id of the workflow to execute (e.g. `uninstall`)
     """
-    if tenant_name:
-        logger.info('Explicitly using tenant `{0}`'.format(tenant_name))
+    utils.explicit_tenant_name_message(tenant_name, logger)
     events_logger = get_events_logger(json_output)
-    events_message = "* Run 'cfy events list -e {0}' to retrieve the " \
+    events_message = "* Run 'cfy events list {0}' to retrieve the " \
                      "execution's events/logs"
     original_timeout = timeout
-    logger.info('Executing workflow {0} on deployment {1} '
-                '[timeout={2} seconds]'.format(
-                    workflow_id,
-                    deployment_id,
-                    timeout))
+    logger.info('Executing workflow `{0}` on deployment `{1}`'
+                ' [timeout={2} seconds]'.format(workflow_id,
+                                                deployment_id,
+                                                timeout))
     try:
         try:
             execution = client.executions.start(
@@ -194,7 +217,11 @@ def manager_start(workflow_id,
                 workflow_id,
                 parameters=parameters,
                 allow_custom_parameters=allow_custom_parameters,
-                force=force)
+                force=force,
+                dry_run=dry_run,
+                queue=queue,
+                wait_after_fail=wait_after_fail,
+                schedule=schedule)
         except (exceptions.DeploymentEnvironmentCreationInProgressError,
                 exceptions.DeploymentEnvironmentCreationPendingError) as e:
             # wait for deployment environment creation workflow
@@ -211,8 +238,8 @@ def manager_start(workflow_id,
                          'workflow execution to finish...')
             now = time.time()
             wait_for_execution(client,
-                               _get_deployment_environment_creation_execution(
-                                   client, deployment_id),
+                               get_deployment_environment_execution(
+                                   client, deployment_id, CREATE_DEPLOYMENT),
                                events_handler=events_logger,
                                include_logs=include_logs,
                                timeout=timeout)
@@ -224,8 +251,19 @@ def manager_start(workflow_id,
                 workflow_id,
                 parameters=parameters,
                 allow_custom_parameters=allow_custom_parameters,
-                force=force)
+                force=force,
+                dry_run=dry_run,
+                queue=queue,
+                wait_after_fail=wait_after_fail,
+                schedule=schedule)
 
+        if execution.status == 'queued':  # We don't need to wait for execution
+            logger.info('Execution is being queued. It will automatically'
+                        ' start when possible.')
+            return
+        if execution.status == 'scheduled':
+            logger.info('Execution is scheduled for {0}.'.format(schedule))
+            return
         execution = wait_for_execution(client,
                                        execution,
                                        events_handler=events_logger,
@@ -268,7 +306,7 @@ def manager_start(workflow_id,
                 workflow_id, deployment_id, e.execution_id, original_timeout))
 
         events_tail_message = "* Run 'cfy events list --tail " \
-                              "--execution-id {0}' to retrieve the " \
+                              "{0}' to retrieve the " \
                               "execution's events/logs"
         logger.info(events_tail_message.format(e.execution_id))
         raise SuppressedCloudifyCliError()
@@ -277,48 +315,111 @@ def manager_start(workflow_id,
 @cfy.command(name='cancel',
              short_help='Cancel a workflow execution [manager only]')
 @cfy.argument('execution-id')
+@cfy.options.common_options
 @cfy.options.force(help=helptexts.FORCE_CANCEL_EXECUTION)
-@cfy.options.verbose()
+@cfy.options.kill()
 @cfy.options.tenant_name(required=False, resource_name_for_help='execution')
 @cfy.assert_manager_active()
 @cfy.pass_client()
 @cfy.pass_logger
-def manager_cancel(execution_id, force, logger, client, tenant_name):
+def manager_cancel(execution_id, force, kill, logger, client, tenant_name):
     """Cancel a workflow's execution
 
     `EXECUTION_ID` is the ID of the execution to cancel.
     """
-    if tenant_name:
-        logger.info('Explicitly using tenant `{0}`'.format(tenant_name))
-    logger.info('{0}Cancelling execution {1}'.format(
-        'Force-' if force else '', execution_id))
-    client.executions.cancel(execution_id, force)
+    utils.explicit_tenant_name_message(tenant_name, logger)
+    if kill:
+        message = 'Killing'
+    elif force:
+        message = 'Force-cancelling'
+    else:
+        message = 'Cancelling'
+    logger.info('{0} execution {1}'.format(message, execution_id))
+    client.executions.cancel(execution_id, force=force, kill=kill)
     logger.info(
         "A cancel request for execution {0} has been sent. "
         "To track the execution's status, use:\n"
         "cfy executions get {0}".format(execution_id))
 
 
-def _get_deployment_environment_creation_execution(client, deployment_id):
-    executions = client.executions.list(deployment_id=deployment_id)
-    for execution in executions:
-        if execution.workflow_id == 'create_deployment_environment':
-            return execution
-    raise RuntimeError('Failed to get create_deployment_environment '
-                       'workflow execution.'
-                       'Available executions: {0}'.format(executions))
+@cfy.command(name='resume',
+             short_help='Resume a workflow execution [manager only]')
+@cfy.argument('execution-id')
+@cfy.options.common_options
+@cfy.options.reset_operations
+@cfy.options.tenant_name(required=False, resource_name_for_help='execution')
+@cfy.assert_manager_active()
+@cfy.pass_client()
+@cfy.pass_logger
+def manager_resume(execution_id, reset_operations, logger, client,
+                   tenant_name):
+    """Resume the execution of a workflow in a failed or cancelled state.
+
+    `EXECUTION_ID` is the ID of the execution to resume.
+    The workflow will run again, restoring the tasks graph from the storage,
+    and retrying failed tasks when necessary.
+    If reset-operations is passed, tasks that were started but didn't fail
+    will be retried as well.
+    """
+    utils.explicit_tenant_name_message(tenant_name, logger)
+    logger.info('Resuming execution {0}'.format(execution_id))
+    client.executions.resume(execution_id, force=reset_operations)
+    logger.info(
+        "A resume request for execution {0} has been sent. "
+        "To track the execution's status, use:\n"
+        "cfy executions get {0}".format(execution_id))
+
+
+@cfy.command(name='list',
+             short_help='List deployment executions')
+@cfy.options.blueprint_id(required=True)
+@cfy.options.common_options
+@cfy.pass_logger
+def local_list(blueprint_id, logger):
+    """Execute a workflow
+
+    `WORKFLOW_ID` is the id of the workflow to execute (e.g. `uninstall`)
+    """
+    env = local.load_env(blueprint_id)
+    executions = env.storage.get_executions()
+    print_data(LOCAL_EXECUTION_COLUMNS, executions, 'Executions:',
+               labels=EXECUTION_TABLE_LABELS)
+
+
+@cfy.command(name='get',
+             short_help='Retrieve execution information')
+@cfy.argument('execution-id')
+@cfy.options.blueprint_id(required=True)
+@cfy.options.common_options
+@cfy.pass_logger
+def local_get(execution_id, blueprint_id, logger):
+    """Retrieve information for a specific execution
+
+    `EXECUTION_ID` is the execution to get information on.
+    """
+    env = local.load_env(blueprint_id)
+    execution = env.storage.get_execution(execution_id)
+    if not execution:
+        raise CloudifyCliError('Execution {0} not found'.format(execution_id))
+    columns = LOCAL_EXECUTION_COLUMNS
+    if get_global_json_output():
+        columns += ['parameters']
+    print_single(LOCAL_EXECUTION_COLUMNS, execution, 'Execution:',
+                 labels=EXECUTION_TABLE_LABELS)
+    if not get_global_json_output():
+        print_details(execution['parameters'], 'Execution Parameters:')
 
 
 @cfy.command(name='start',
              short_help='Execute a workflow')
 @cfy.argument('workflow-id')
-@cfy.options.blueprint_id(required=True, multiple_blueprints=True)
+@cfy.options.blueprint_id(required=True)
 @cfy.options.parameters
 @cfy.options.allow_custom_parameters
 @cfy.options.task_retries()
 @cfy.options.task_retry_interval()
 @cfy.options.task_thread_pool_size()
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.pass_logger
 def local_start(workflow_id,
                 blueprint_id,
@@ -341,3 +442,87 @@ def local_start(workflow_id,
                          task_thread_pool_size=task_thread_pool_size)
     if result is not None:
         logger.info(json.dumps(result, sort_keys=True, indent=2))
+
+
+@executions.command(name='summary',
+                    short_help='Retrieve summary of execution details '
+                               '[manager only]')
+@cfy.argument('target_field', type=click.Choice(EXECUTIONS_SUMMARY_FIELDS))
+@cfy.argument('sub_field', type=click.Choice(EXECUTIONS_SUMMARY_FIELDS),
+              default=None, required=False)
+@cfy.options.common_options
+@cfy.options.tenant_name(required=False, resource_name_for_help='summary')
+@cfy.options.all_tenants
+@cfy.pass_logger
+@cfy.pass_client()
+def summary(target_field, sub_field, logger, client, tenant_name,
+            all_tenants):
+    """Retrieve summary of executions, e.g. a count of each execution with
+    the same deployment ID.
+
+    `TARGET_FIELD` is the field to summarise executions on.
+    """
+    utils.explicit_tenant_name_message(tenant_name, logger)
+    logger.info('Retrieving summary of executions on field {field}'.format(
+        field=target_field))
+
+    summary = client.summary.executions.get(
+        _target_field=target_field,
+        _sub_field=sub_field,
+        _all_tenants=all_tenants,
+    )
+
+    columns, items = structure_summary_results(
+        summary.items,
+        target_field,
+        sub_field,
+        'executions',
+    )
+
+    print_data(
+        columns,
+        items,
+        'Execution summary by {field}'.format(field=target_field),
+    )
+
+
+@executions.command(name='delete',
+                    short_help='Delete finished executions')
+@cfy.options.common_options
+@cfy.options.tenant_name(required=False,
+                         resource_name_for_help='executions')
+@cfy.options.to_datetime(required=False,
+                         mutually_exclusive_with=['before', 'keep_last'],
+                         help="Executions that were created at this timestamp"
+                              " or before will be deleted")
+@cfy.options.before(required=False,
+                    mutually_exclusive_with=['to_datetime', 'keep_last'],
+                    help="Executions that were created this long ago or "
+                         "earlier will be deleted (e.g. '2 weeks')")
+@cfy.options.keep_last("executions",
+                       required=False,
+                       mutually_exclusive_with=['before', 'to_datetime'])
+@cfy.options.all_tenants
+@cfy.pass_logger
+@cfy.pass_client()
+def delete(logger, client, tenant_name, to_datetime, before, keep_last,
+           all_tenants):
+    """Delete executions from the executions list, by specifying a number of
+    executions to keep, a number of days to keep executions for, or a date
+    starting from which to keep executions.
+
+    * Only deletes finished executions, i.e. completed, failed or cancelled.
+
+    * Does not delete the latest deployment environment creation for each
+    deployment."""
+    if before:
+        to_datetime = before
+
+    utils.explicit_tenant_name_message(tenant_name, logger)
+    deleted_executions_count = client.executions.delete(
+        to_datetime=to_datetime, keep_last=keep_last, _all_tenants=all_tenants)
+    deleted_executions_count = deleted_executions_count
+    if deleted_executions_count:
+        logger.info('\nDeleted %d executions', deleted_executions_count)
+    else:
+        logger.info('\nNo executions to delete')
