@@ -42,12 +42,16 @@ class ExecutionEventsFetcher(object):
                  client,
                  execution_id,
                  batch_size=100,
-                 include_logs=False):
+                 include_logs=False,
+                 from_datetime=None,
+                 to_datetime=None):
         self._client = client
         self._execution_id = execution_id
         self._batch_size = batch_size
         self._from_event = 0
         self._include_logs = include_logs
+        self._from_datetime = from_datetime
+        self._to_datetime = to_datetime
         # make sure execution exists before proceeding
         # a 404 will be raised otherwise
         self._client.executions.get(execution_id)
@@ -75,7 +79,9 @@ class ExecutionEventsFetcher(object):
             _offset=offset,
             _size=size,
             include_logs=self._include_logs,
-            sort='reported_timestamp')
+            sort='reported_timestamp',
+            from_datetime=self._from_datetime,
+            to_datetime=self._to_datetime)
         self._from_event += len(events_list_response)
         return events_list_response
 
@@ -102,6 +108,9 @@ class ExecutionEventsFetcher(object):
             del event[context_field]
 
         event['context']['node_id'] = event['node_instance_id']
+        if 'source_id' in event:
+            event['context']['source_id'] = event['source_id']
+            event['context']['target_id'] = event['target_id']
         del event['node_instance_id']
 
         event['message'] = {
@@ -182,7 +191,8 @@ def wait_for_execution(client,
                        events_handler=None,
                        include_logs=False,
                        timeout=900,
-                       logger=None):
+                       logger=None,
+                       from_datetime=None):
 
     # if execution already ended - return without waiting
     if execution.status in Execution.END_STATES:
@@ -193,7 +203,8 @@ def wait_for_execution(client,
 
     events_fetcher = ExecutionEventsFetcher(client,
                                             execution.id,
-                                            include_logs=include_logs)
+                                            include_logs=include_logs,
+                                            from_datetime=from_datetime)
 
     # Poll for execution status and execution logs, until execution ends
     # and we receive an event of type in WORKFLOW_END_TYPES
@@ -220,10 +231,8 @@ def wait_for_execution(client,
             execution = client.executions.get(execution.id)
             execution_ended = execution.status in Execution.END_STATES
 
-        if not events_watcher.end_log_received and \
-                execution.status != Execution.PENDING:
-            events_fetcher.fetch_and_process_events(
-                events_handler=events_watcher, timeout=timeout)
+        events_fetcher.fetch_and_process_events(
+            events_handler=events_watcher, timeout=timeout)
 
         if execution_ended and events_watcher.end_log_received:
             break
