@@ -16,6 +16,8 @@
 
 import os
 import sys
+import json
+import click
 import errno
 import string
 import random
@@ -23,21 +25,38 @@ import shutil
 import tarfile
 import zipfile
 import tempfile
+import collections
 from shutil import copy
-from urlparse import urlparse
 from contextlib import closing, contextmanager
 from backports.shutil_get_terminal_size import get_terminal_size
 
+import yaml
 import requests
 
 from .logger import get_logger
 from .exceptions import CloudifyCliError
 from .constants import SUPPORTED_ARCHIVE_TYPES
 
+from cloudify._compat import urlparse
 from cloudify_rest_client.constants import VisibilityState
 from cloudify_rest_client.exceptions import CloudifyClientError
 
 
+def get_deployment_environment_execution(client, deployment_id, workflow):
+    executions = client.executions.list(deployment_id=deployment_id,
+                                        workflow_id=workflow,
+                                        sort='created_at',
+                                        is_descending=True)
+
+    if executions and len(executions) > 0:
+        return executions[0]
+
+    raise RuntimeError(
+        'Failed to get {0} workflow execution for deployment {1}'.format(
+            workflow, deployment_id)
+    )
+
+
 def dump_to_file(collection, file_path):
     with open(file_path, 'a') as f:
         f.write(os.linesep.join(collection))
@@ -45,6 +64,9 @@ def dump_to_file(collection, file_path):
 
 
 def is_virtual_env():
+    if hasattr(sys, 'base_prefix'):
+        # py3 case, with the stdlib venv
+        return sys.base_prefix != sys.prefix
     return hasattr(sys, 'real_prefix')
 
 
@@ -55,34 +77,6 @@ def get_cwd():
     return os.getcwd()
 
 
-def decode_list(data):
-    rv = []
-    for item in data:
-        if isinstance(item, unicode):
-            item = item.encode('utf-8')
-        elif isinstance(item, list):
-            item = decode_list(item)
-        elif isinstance(item, dict):
-            item = decode_dict(item)
-        rv.append(item)
-    return rv
-
-
-def decode_dict(data):
-    rv = {}
-    for key, value in data.iteritems():
-        if isinstance(key, unicode):
-            key = key.encode('utf-8')
-        if isinstance(value, unicode):
-            value = value.encode('utf-8')
-        elif isinstance(value, list):
-            value = decode_list(value)
-        elif isinstance(value, dict):
-            value = decode_dict(value)
-        rv[key] = value
-    return rv
-
-
 def remove_if_exists(path):
 
     try:
@@ -144,12 +138,12 @@ def zip_files(files):
     destination_zip = source_folder + '.zip'
     for path in files:
         copy(path, source_folder)
-    zip(source_folder, destination_zip, include_folder=False)
+    create_zip(source_folder, destination_zip, include_folder=False)
     shutil.rmtree(source_folder)
     return destination_zip
 
 
-def zip(source, destination, include_folder=True):
+def create_zip(source, destination, include_folder=True):
     logger = get_logger()
 
     logger.debug('Creating zip archive: {0}...'.format(destination))
@@ -240,8 +234,8 @@ def generate_progress_handler(file_path, action='', max_bar_length=80):
 
     # Shorten the file name if it's too long
     file_name = os.path.basename(file_path)
-    if len(file_name) > (bar_length / 4) + 3:
-        file_name = file_name[:bar_length / 4] + '...'
+    if len(file_name) > (bar_length // 4) + 3:
+        file_name = file_name[:bar_length // 4] + '...'
 
     bar_length -= len(file_name)
 
@@ -262,8 +256,8 @@ def generate_progress_handler(file_path, action='', max_bar_length=80):
 
         # The \r caret makes sure the cursor moves back to the beginning of
         # the line
-        sys.stdout.write('\r{0} {1} |{2}| {3}%'.format(
-            action, file_name, bar, percents))
+        msg = '\r{0} {1} |{2}| {3}%'.format(action, file_name, bar, percents)
+        click.echo(msg, nl=False)
         if read_bytes >= total_bytes:
             sys.stdout.write('\n')
 
@@ -276,7 +270,7 @@ def handle_client_error(status_code, message, logger):
     """
     try:
         yield
-    except CloudifyClientError, e:
+    except CloudifyClientError as e:
         if e.status_code != status_code:
             raise
         logger.info(message)
@@ -291,7 +285,7 @@ def prettify_client_error(status_codes, logger):
     """
     try:
         yield
-    except CloudifyClientError, e:
+    except CloudifyClientError as e:
         if e.status_code not in status_codes:
             raise
         logger.info(e.message)
@@ -312,7 +306,7 @@ def get_visibility(private_resource,
 
 
 def validate_visibility(visibility, valid_values=VisibilityState.STATES):
-    if visibility not in valid_values:
+    if visibility and visibility not in valid_values:
         raise CloudifyCliError(
             "Invalid visibility: `{0}`. Valid visibility's values are: "
             "{1}".format(visibility, valid_values)
@@ -337,3 +331,76 @@ def get_local_path(source, destination=None, create_temp=False):
         raise CloudifyCliError(
             'You must provide either a path to a local file, or a remote URL '
             'using one of the allowed schemes: {0}'.format(allowed_schemes))
+
+
+def explicit_tenant_name_message(tenant_name, logger):
+    if tenant_name:
+        logger.info('Explicitly using tenant `{0}`'.format(tenant_name))
+
+
+def deep_update_dict(dest_dict, src_dict):
+    for key, value in src_dict.items():
+        if isinstance(dest_dict, collections.MutableMapping):
+            if isinstance(value, collections.MutableMapping):
+                dest_dict[key] = deep_update_dict(dest_dict.get(key), value)
+            else:
+                dest_dict[key] = src_dict[key]
+        else:
+            dest_dict = {key: src_dict[key]}
+    return dest_dict
+
+
+def deep_subtract_dict(dest_dict, src_dict):
+    for key, value in src_dict.items():
+        if isinstance(value, collections.MutableMapping):
+            deep_subtract_dict(dest_dict.get(key), value)
+        else:
+            if key not in dest_dict:
+                raise CloudifyCliError('Key {} does not exist'.format(key))
+            dest_dict.pop(key)
+
+
+def insert_dotted_key_to_dict(dest_dict, key, value):
+    """Insert the value into dest_dict according to the key which is in dot
+       hierarchy format
+
+    :param dest_dict: The dict to update
+    :param key: The dot hierarchy key, e.g. 'a.b.c'
+    :param value: The value to insert, e.g. 'd'
+    :return: dest_dict will include the value in the wanted location,
+        e.g. {a: {b: {c: d}}}
+    """
+    key_path = key.split('.')
+    for item in key_path[:-1]:
+        dest_dict.setdefault(item, {})
+        dest_dict = dest_dict[item]
+    dest_dict.setdefault(key_path[-1], value)
+
+
+def assert_one_argument(arguments):
+    """Asserts exactly one argument in a dictionary of
+    {argument_name: argument} is not null or False, else raises an error
+    """
+    filtered = [k for k in arguments if arguments[k]]
+    if len(filtered) != 1:
+        raise CloudifyCliError('Please provide one of the options: ' +
+                               ', '.join(arguments))
+
+
+def load_json(input_path):
+    if not input_path:
+        return
+    with open(input_path) as json_file:
+        return json.load(json_file)
+    # return json_content
+
+
+def print_dict(keys_dict, logger):
+    for key, values in keys_dict.items():
+        str_values = [str(value) for value in values]
+        logger.info('{0}: {1}'. format(key, str_values))
+
+
+def get_dict_from_yaml(yaml_path):
+    with open(yaml_path) as f:
+        return yaml.load(f, yaml.Loader)
